# -*- coding: utf-8 -*-

# Copyright 2006-2012 Borislav Rizov <bobyATdclDOTbasDOTbg>, Department of Computational Linguistic, Institute for Bulgarian Language – Bulgarian Academy of Sciences
# 
# This file is part of Hydra.
# 
# Hydra is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hydra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Hydra.  If not, see <http://www.gnu.org/licenses/>.

import MySQLdb
import re, types
import datetime
import settings
import getpass


class DB:
    db = settings.db
    host = settings.host
    charset = settings.encoding
    user = settings.user
    passwd = settings.password
    autocommit = True
    
    def __init__(self):
        self.connect()

    def _connect(self):
        self.conn = MySQLdb.connect(db = self.db, host = self.host, user = self.user, passwd = self.passwd, charset = self.charset, use_unicode=True)
        self.conn.autocommit(self.autocommit)

    def connect(self):
        disconnected = True
        while disconnected:
            try:
                self._connect()
                disconnected = False
            except MySQLdb.OperationalError, e:
                if e.args[0] in (1044, 1045):
                    self.read_user_pass()
                elif e.args[0] in (1049, 2003):
                    self.read_host_db()
                else:
                    print e
        self.post_connect()
    
    def db_name(self):
        return '%s@%s' % (self.db, self.host)
    
    def raw_input(self, prompt, value):
        if value:
            new_value = raw_input('%s (%s): ' % (prompt, value))
            if not new_value:
                new_value = value
        else:
            new_value = raw_input('%s' % prompt)
        return new_value

    def read_user_pass(self):
        self.user = self.raw_input('user', self.user)
        self.passwd = getpass.getpass(prompt = 'password: ')
    
    def read_host_db(self):
        self.host = self.raw_input('host', self.host)
        self.db = self.raw_input('db', self.db)
    
    def post_connect(self):
        cursor = self.conn.cursor()
        cursor.execute(u"select user()")
        self.username = cursor.fetchone()[0]
#        print self.user

    def _execute(self, query):
        cursor = self.conn.cursor()
        cursor.execute(query)
        return cursor

    def execute(self, query):
        if  type(query) is types.StringType:
            query = query.decode(settings.encoding)
        try:
            return self._execute(query)
        except (AttributeError, MySQLdb.OperationalError):
            self._connect()
            return self._execute(query)

    def create_vals(self, values):
        vals = '(' + ", ".join(map(self.my_q, values)) + ')'
        return vals
    
    def insert(self, table, values):
        vals = self.create_vals(values)
        query = "insert ignore into %s values %s;" % (table, vals)
        self.execute(query)
    
    def last_id(self):
        cursor = self.execute("select last_insert_id();")
        return long(cursor.fetchone()[0])
    
    def replace(self, table, values):
        vals = self.create_vals(values)
        query = "replace into %s values %s;" % (table, vals)
        #print type(query), "replace type"
        self.execute(query)
        
    def my_q(self, string):
# 		print type(string)
        if string == None:
            return "NULL"
        if( type(string) is types.UnicodeType):
            string  = string.encode(settings.encoding)
        elif type(string) is datetime.datetime:
            string = str(string)
            #print string
        
        if type(string) is types.StringType:
            string = self.conn.escape_string(string)
            return "'" + string + "'"
        return str(string)
    
    def get_one(self, id, table):
        query = "select * from %s where id = \"%s\"" %(table, id)
        cursor = self.execute(query)
        res = cursor.fetchone()
        return res
    
    def get_all(self, table):
        query = "select * from %s" % table
        cursor = self.execute(query)
        return cursor.fetchall()
    
    def get_values(self, field, table):
        query = "select * from %s" % table
        cursor = self.execute(query)
        return [r[field] for r in cursor.fetchall()]

    def get_server_timestamp(self):
        query = "select now();"
        cursor = self.execute(query)
        return cursor.fetchone()[0]
    
    def get_user(self):
        return self.username

    def find(self, field, value, table):
        query = "select * from %s where %s = %s" %(table, field, self.my_q(value))
##      print query
        cursor = self.execute(query)
        res = cursor.fetchone()
        return res
    
    def findall(self, field, value, table):
        query = "select * from %s where %s = %s" %(table, field, self.my_q(value))
##      print query
        cursor = self.execute(query)
        res = cursor.fetchall()
        return res

class CachedTable(object):
    'Assuming that the string representation is row[1]'
    def __init__(self, db, table):
        self.db = db
        self.table = table
        self.data = None

    def load(self):
        if self.data == None:
           self._load()

    def _load(self):
        data = self.db.get_all(self.table)
        self.data = {}
        self.index = {}
        for row in data:
            self.add_row(row)
    
    def add_row(self, row):
        self.data[row[0]] = row
        self.index[row[1]] = row


    def from_string(self, string):
        self.load()
        return self.index.get(string, None)
    
    def to_string(self, id):
        self.load()
        row = self.from_id(id)
        if row:
            return row[1]
    
    def to_id(self, string):
        data = self.from_string(string)
        if data:
            return data[0]
    
    def from_id(self, id):
        self.load()
        return self.data.get(id, None)
    
    def string_values(self):
        self.load()
        return [self.to_string(x) for x in self.data]
   
class DCachedTable(CachedTable):
    def __init__(self, db, descriptor):
        CachedTable.__init__(self, db, descriptor.table)
        self.descriptor = descriptor

    def add(self, string, **kwargs):
        string = string.strip()
        row = self.from_string(string)
        if row:
            return row
        else:
            values = ["'%s'" % string] + ['DEFAULT'] * (len(self.descriptor.fields) - 2)
#            print 'kw', kwargs, values
            for key, value in kwargs.iteritems():
                values[self.descriptor[key] - 1] = "'%s'" % value
            values = ', '.join(values)
            query = "insert into %s values (0, %s)" % (self.table, values)
#            print query
            self.db.execute(query)
            rid = self.db.last_id()
            row = self.db.get_one(rid, self.table)
            self.add_row(row)
            return row
    
    def failsafe_add(self, value):
        if value:
            return self.add(value)[self.descriptor.id]
        return 0
        
    
class CachedTablePool(object):
    def __init__(self, db):
        self.db = db
        self.tables = {}
    
    def get(self, table):
        if table not in self.tables:
            self.tables[table] = CachedTable(self.db, table)
        return self.tables[table]
    
    def __getitem__(self, table):
        return self.get(table)


if __name__ == '__main__':
    def test():
        import psycopg2
        connection = psycopg2.connect("dbname='wn_test' user='boby' host='localhost' password='boza'")
        cursor = connection.cursor()
        cursor.execute('select user')
        print cursor.fetchone()
    test()
    
