# -*- coding: utf-8 -*-

# Copyright 2006-2012 Borislav Rizov <bobyATdclDOTbasDOTbg>, Department of Computational Linguistic, Institute for Bulgarian Language – Bulgarian Academy of Sciences
# 
# This file is part of Hydra.
# 
# Hydra is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hydra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Hydra.  If not, see <http://www.gnu.org/licenses/>.

import MySQLdb
from xml_import import *
from relations import *
import sys
from xml.dom import minidom

class DomXmlParser(object):
    def search(self, tag, element): 
        result = element.getElementsByTagName(tag)
        if result:
            return self.text(result[0])
        
    def all(self, tag, element):
        return element.getElementsByTagName(tag)
        
    def text(self, element):
        rc = []
        for node in element.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc.append(node.data)
        return ''.join(rc)


class DBFeeder(DB):
    autocommit = False
    engine = 'InnoDB'
    langs_query = """
        create table if not exists Lang(
        id int unsigned not null auto_increment primary key,
        name char(30) unique not null,
        synset_id int not null
        )
    """
    rel_query = """
        create table if not exists Rel(
        id1 bigint unsigned not null,
        id2 bigint unsigned not null, 
        rel int(2) unsigned not null,
        index index1 (id1), 
        index index2 (id2), 
        index rel_index (rel)
        ) charset utf8;
    """
    literal_query = """
        create table if not exists Literal(
        id bigint unsigned not null auto_increment primary key,
        sense int unsigned not null,
        word varchar(200) not null,
        lemma varchar(200) not null,
        index word_index (word)
        ) charset utf8;"""
    synset_query = """
        create table if not exists Synset(
        id bigint unsigned not null primary key,
        ili char(20) not null,
        pos int(2) not null,
        definition varchar(4000) not null,
        stamp char(60),
        bcs int(2) unsigned not null default 0,
        lang int(3) unsigned not null,
        frequency int not null default 0,
        domain int(4) not null,
        sem_class int(4) not null,

        index (ili),
        index (pos),
        index (definition),
        index (stamp),
        index (bcs),
        index (lang),
        index (frequency),
        index (domain),
        index (sem_class)
        ) charset utf8;"""
    relations_query = """
        create table if not exists Relation(
        id int(2) unsigned not null auto_increment primary key,
        name char(30) unique,
        rname char(30) default '',
        id1 char(10) default 'Synset', 
        id2 char(10) default 'Synset',
        characteristics int not null default 0
        )""" 
    notes_query = """
        create table if not exists Note(
        id bigint unsigned not null auto_increment primary key,
        note varchar(1000) not null
        ) charset utf8, auto_increment = 10000000;
    """
    pos_query = """
        create table if not exists POS(
        id int unsigned not null auto_increment primary key,
        name char(8) unique not null
        ) charset utf8;
    """
    domain_query = """
        create table if not exists Domain(
        id int unsigned not null auto_increment primary key,
        name char(64) unique not null
        ) charset utf8;
    """
    sem_class_query = """
        create table if not exists SemClass(
        id int unsigned not null auto_increment primary key,
        name char(64) unique not null,
        code char(4)
        ) charset utf8;
    """

    lock_query = """
        create table if not exists Locks (
        id bigint unsigned not null auto_increment primary key,
        stamp timestamp, 
        user varchar(100), 
        value int unsigned not null default 0 
        ) engine = memory;
    """

    def __init__(self):
        DB.__init__(self)

        def create():
            self.execute('set storage_engine=%s' % self.engine)
            self.execute(self.langs_query)
            self.execute(self.lock_query)
            self.execute(self.synset_query)
            self.execute(self.literal_query)
            self.execute(self.rel_query)
            self.execute(self.relations_query)
            self.execute(self.notes_query)
            self.execute(self.pos_query)
            self.execute(self.domain_query)
            self.execute(self.sem_class_query)

        def fill():
            self.fill_rels()
            self.load_senses()
            self.fill_pos()
            self.fill_domain()
            self.conn.commit()
        
        create()
        fill()
        self.init()
        self.ids = {}
        self.all_ids = {}
   
    def init(self):
        from wordnet import wn
        self.wn = wn
        self.pos = DCachedTable(self, POSId)
        self.rel = DCachedTable(self, RelationId)
        self.domain = DCachedTable(self, DomainId)
        self.sem_class = DCachedTable(self, SemClassId)

    def info(self, filename, ln, delim = '\r'):
        print filename + " "*(80 - len(filename) - len(ln)) + ln + delim,

    
    def fill_rels(self):
        query = """
            insert ignore into %(table)s values 
            (0, 'literal', 'synset', 'Synset', 'Literal', %(nc_ex)d),
            (0, 'snote', '', 'Synset', 'Note', %(nc_ex)d),
            (0, 'lnote', '', 'Literal', 'Note', %(nc_ex)d),
            (0, 'usage', '', 'Synset', 'Note', %(nc_ex)d),
            (0, 'ili', '', 'Synset', 'Synset', %(system)d),
            (0, 'hypernym', 'hyponym', 'Synset', 'Synset', %(tr)d),
            (0, 'instance_hypernym', 'instance_hyponym', 'Synset', 'Synset', %(tr)d),
            (0, 'holo_part', 'mero_part', 'Synset', 'Synset', %(tr)d),
            (0, 'holo_member', 'mero_member', 'Synset', 'Synset', %(tr)d),
            (0, 'holo_portion', 'mero_portion', 'Synset', 'Synset', %(tr)d),
            (0, 'near_antonym', 'near_antonym', 'Synset', 'Synset', %(sym)d),
            (0, 'also_see', 'also_see', 'Synset', 'Synset', %(sym)d),
            (0, 'bg_derivative', 'bg_derivative', 'Synset', 'Synset', %(sym)d),
            (0, 'eng_derivative', 'eng_derivative', 'Synset', 'Synset', %(sym)d),
            (0, 'similar_to', 'similar_to', 'Synset', 'Synset', %(sym)d),
            (0, 'verb_group', 'verb_group', 'Synset', 'Synset', %(sym)d),
            (0, 'subevent', 'is_subevent_of', 'Synset', 'Synset', %(tr)d),
            (0, 'causes', 'is_caused_by', 'Synset', 'Synset', %(ord)d),
            (0, 'derived', 'is_derived_from', 'Synset', 'Synset', %(ord)d),
            (0, 'be_in_state', 'is_state_of', 'Synset', 'Synset', %(ord)d),
            (0, 'participle', 'is_participle_of', 'Synset', 'Synset', %(ord)d),
            (0, 'category_domain', 'category_member', 'Synset', 'Synset', %(ord)d),
            (0, 'region_domain', 'region_member', 'Synset', 'Synset', %(ord)d),
            (0, 'usage_domain', 'usage_member', 'Synset', 'Synset', %(ord)d)
        """ %{'table': RelationId.table, 'nc_ex':RelChars.NewCreational | RelChars.Explicit, 'tr':RelChars.Transitive, 'system':RelChars.System, 'sym':RelChars.Symetric, 'ord':0}
        self.execute(query)
        
    def fill_pos(self):
        query = "insert ignore into %s values (0, 'a'), (0, 'b'), (0, 'n'), (0, 'v'), (0, 'p'), \
        (0, 'pron'), (0, 'conj'), (0, 'particle'), (0, 'ij')" % POSId.table
        self.execute(query)

    def fill_domain(self):
        query = "insert ignore into %s values (0, 'chemistry'), (0, 'news'), (0, 'law')" % DomainId.table
        self.execute(query)

    def fill_ili(self):
        query = """insert ignore into %(rel)s select T1.id as id1, T2.id as id2, (select id from %(relations)s where name = 'ili') as rel 
        from %(main)s as T1 inner join %(main)s  as T2 on T1.ili = T2.ili and T1.id != T2.id""" % {'rel': RelId.table, 'relations': RelationId.table, 'main': SynsetId.table}
        self.execute(query)

    def get_rel(self, name, id1='Synset', id2='Synset', **kwargs):
        name = name.lower()
        rel = self.rel.add(name, id1 = id1, id2 = id2, **kwargs)
        return rel[RelationId.id]
    
    def load_senses(self):
        self.senses = {}
        cursor = self.execute('select word, max(sense) from Literal group by word')
        word_senses = cursor.fetchall()
        for ws in word_senses:
            self.senses[ws[0]] = ws[1]
    
    def norm_ili(self,  ili,  lang):
        parts = ili.split('-')
        if len(parts) == 2:
            suffix = 1
        else:
            suffix = -2
        return '-'.join([lang['name']] +  parts[suffix:])
    
    def encode(self,  ili,  lang):
        ili = self.norm_ili(ili,  lang)
        self.ids[ili] = self.ids.get(ili,  -1) + 1
        id =  wn.encode(ili,  lang,  self.ids[ili])
        if id in self.all_ids:
            print ili,  self.all_ids[id]
        self.all_ids[id] = ili
        return id
        
    def add_rel(self, id1, ili2, rel_name, lang):
        type1 = self.wn.ling(id1).__class__.__name__
        if ':' in ili2:
            type2 = 'Literal'
            relid = self.get_rel(rel_name, type1, type2)
            self.postponed.append((id1, ili2, relid))
        else:
            type2 = 'Synset'
            relid = self.get_rel(rel_name, type1, type2)
            try:
                id2 = wn.encode(ili2, lang)
                self.insert(RelId.table, (id1, id2, relid))
            except:
                print 'problematic', ili2

    def add_postponed(self, lang):
        for id1, ili2, relid in self.postponed:
            ili, index = ili2.split(':', 1)
            index = int(index)
            id2 = self.wn.encode(ili, lang)
            synset = Synset(long(id2))
            literals = synset.literals()
            literals.sort(key = lambda literal: literal.ID())
            print len(literals), index, ili2, id2, type(id2)
            lid = literals[index].ID()
            self.insert(RelId.table, (id1, lid, relid))

    def add_lang(self,  name):
        return self.wn.add_lang(name)
    
    def fix_lang(self, lang, name):
        query = 'select max(ili) from %s where ili like "%s-%%"' % (SynsetId.table, name)
        cursor = self.execute(query)
        max = cursor.fetchone()[0]
        if max:
            max = long(max.split('-')[1]) + 1
            cursor = self.execute('update Lang set synset_id=%s where id=%s' % (max, lang))

    def get_sense(self, word):
        sense = self.senses.get(word, 0) + 1
        self.senses[word] = sense
        return sense

    def add_note(self, xid, text, rel):
        self.insert(NoteId.table, (0, text))
        noteid = self.last_id()
        self.insert(RelId.table, (xid, noteid, rel))


class LineDBFeeder(DBFeeder):

    def init(self):
        DBFeeder.init(self)
        self.splits = set()
        self.problem = 0


    def feed(self, wns):
        for _wn in wns:
            self.postponed = []
            lang = self.add_lang(_wn[0])
            self.load(_wn[1], lang)
            self.fix_lang(lang.ID(), _wn[0])
            self.add_postponed(lang)
        self.problems(sys.stderr)
        self.fill_ili()
        self.conn.commit()

    def problems(self,  output):
        for ili,  count in self.ids.iteritems():
            if count > 0:
                print >> output,  ili[5:]
        print >> output,  'splits:'
        for s in self.splits:
            print >> output,  s

    def add(self, synset, lang):
        wn = self.wn
        ili = self.im.search("ID", synset)
        if not ili: 
            print 'problem with %s' % synset
            return
        pos_name = self.im.search("POS", synset)
        pos = self.pos.failsafe_add(pos_name)
        domain_name = self.im.search("DOMAIN", synset)
        domain = self.domain.failsafe_add(domain_name)
        bcs = self.im.search("BCS", synset)
        stamp = self.im.search("STAMP", synset)
        defin = self.im.search("DEF", synset)
        frequency = self.im.search("FREQUENCY",  synset)

        sem_clas_name = self.im.search("SEM_CLASS", synset)
        sem_class = self.sem_class.failsafe_add(sem_clas_name)

        id = self.encode(ili,  lang)
        self.insert(SynsetId.table, (id, ili, pos, defin, stamp, bcs, lang['id'],  frequency, domain, sem_class))
        #lits
        lits = self.im.all("LITERAL", synset)
        lit_rel = self.get_rel('literal')
        for lit in lits:
            lid = self.add_literal(lit, lang)
            self.insert(RelId.table, (id, lid, lit_rel))
        #ilrs
        ilrs = self.im.all("ILR", synset)
        for ilr in ilrs:
            rel_name = self.im.search("TYPE", ilr)
            if rel_name.strip() == '':
                print synset
            if rel_name == 'SPLIT':
                self.splits.add(id)
            ili2 = self.im.text(ilr)
#            print rel_name,  ili2,  synset
            self.add_rel(id, ili2, rel_name, lang)
        self.add_notes(id, "SNOTE", "snote", synset)
        self.add_notes(id, "USAGE", "usage", synset)
    
    def add_literal(self, lit, lang):
        lemma = self.im.search("LEMMA", lit)
        word = self.im.text(lit)
        sense = self.get_sense(word)
        self.insert(LiteralId.table, (0, sense, word, lemma))
        lid = self.last_id()
        self.add_notes(lid, "LNOTE", "lnote", lit)
        lrs = self.im.all('LR', lit)
        for lr in lrs:
            rel_name = self.im.search('TYPE', lr)
            ili2 = self.im.text(lr)
            self.add_rel(lid, ili2, rel_name, lang)
        return lid
    
    def add_notes(self, xid, note_type, rel_name, text):
        rel = self.get_rel(rel_name)
        notes = self.im.all(note_type, text)
        
        for note in notes:
            note_text = self.im.text(note)
            self.add_note(xid, note_text, rel)

    def load(self, filename, lang):
        self.im = XmlParser()
        
        file = open(filename, 'r')
        line = 0
        for synset in file:
            self.add(synset, lang)
            self.info(filename, str(line))
            line += 1
        self.info(filename, "done", '\n')


class DomDBFeeder(LineDBFeeder):
    def load(self, filename, lang):
        self.im = DomXmlParser()
        
        dom = minidom.parse(filename)
        line = 0
        synsets = dom.getElementsByTagName('SYNSET')
        for synset in synsets:
            self.add(synset, lang)
            self.info(filename, str(line))
            line += 1
        print 
        self.info(filename, "done", '\n')

if __name__ == '__main__':
    
    def feed(load_type=None):
        if load_type == 'dom':
            feeder = DomDBFeeder()
        else:
            feeder = LineDBFeeder()

        wns = settings.wordnets
        feeder.feed(wns)

    feed(*sys.argv[1:])

